#!/bin/sh
# backup.sh
# 
# Backup up Unix-like computers
# 
# itk@samfundet.no
#
# This script was first checked into RCS in 2000, and has since backed up a few
# servers almost every day, whenever not broken.
# 
# It's a bit peculiar, but quite effective. The following is worth keeping in
# mind when hacking:
#
# KISS. And by that I mean really simple. Only regular Unix commands, please
# (but GNU extensions are used).
#
# Because of the way we run this script from cron, informational logging goes
# to stderr and warnings and errors to stdout. This way, we get mail whenever
# stuff fails, and other output is redirected to the log file.
# 
# The computers are backed up in parallell (but the filesystems on each
# computer in serial). This is implemented highly simplistic, with forking
# subshells. The script has no concept about threads, and stuff might go wrong,
# which usually requires manual cleanups.
#
# The script is halfway Norwegian and halfway English. Newer modifications are
# in English, we should probably stick to that.
# 
# Known issues:
#
# o $R has no proper meaning because of the forking subshell stuff. We need a
#   more robust way of returning non-zero on errors, which we probably don't
#   right now.
# o The code around line 211 and downwards is messy.
#

# Configuration
LOCKFILE=/home/backup/backuprun.lockfile
confdir=/home/backup/conf/           # Configuration files
storagedir=/backup                   # Where we keep backups
maksantallfulle=3                    # Number of full backups
dagermellomfulle=30                  # Days between full backups
DATE=`date "+%Y%m%d%H%M"`            #format: touch
DATEs=`date "+%Y-%m-%d %H:%M"`       #format: tar

# Exclude-pattern
exclude=$confdir/exclude
[ ! -f $exclude ] && printf "tmp\ncore\n" > $exclude

# Initially, we consider everything to be just fine.
R=0

# Die more or less gracefully (at least notify user)
die() {
	echo `date`": Something nasty happened, and since I fork a lot:"
        echo `date`": I CANNOT CLEAN UP THE MESS MYSELF."
	echo `date`": You need to get rid of lost process named stuff like $0, tar and ssh."
	exit 255
}

# Trap C-c and kill
trap die SIGINT SIGTERM

# Don't start if we're already running
if [ -e $LOCKFILE ]; then
	echo `date`": $LOCKFILE exists, exiting."
	echo `date`": $LOCKFILE exists, exiting." >&2
	exit 1
fi;
touch $LOCKFILE
echo "$$" > $LOCKFILE

#syntax of remotestat:
#hostname:/directory/
remotestatf=$confdir/remotestat
[ -f $remotestatf ] && remotestat=`cat $remotestatf`
###

PATH=/local/bin:$PATH:/store/bin
export PATH

echo `date`": Backup run starting" >&2

umask 027

# The computers we want to back up
unixcomputers=`cat $confdir/computers.unix \
  | grep -v "^#" \
  | grep -v "^$" `

# Backup only one computer, from command line?
if [ $1 ] ; then 
  unixcomputers=$1
fi;


#selve backupen
# krever at noen variabler er satt
# krever at vi er i rett katalog
backup()
{

 echo -n `date` >&2
 printf " $computer: $computer:$filesystem $backuplevel backup" >&2

 if [ "$backuplevel" = "daglig" ] || [ "$backuplevel" = "incremental" ]
 then
  lastd=`cat ../.date`
  lastcmd="--newer='$lastd'"
 else
  lastcmd=""
 fi

 #if this client has a special tar
 #we need to find a better solution to this conf-issue
 if [ -f $confdir/tar.$computer ] ; then
   tar=`cat $confdir/tar.$computer`
 else 
   tar=tar
 fi

 # expand the exclude-path for use with tar
 exf=`ssh root@$computer "ls ~/.backup/exclude"`

 # We try to run tar on the remote computer
 #    z gzip it
 #    c create archive
 #    C change to directory first
 #    . where to start taring (see C)
 #    $lastcmd only files newer than this
 #    --one-file-system don't traverse file systems
 #    --exclude-from file to get exclusion pattern from
 #    Pipe the stuff over ssh to ourselves, run pee to cat the contents to a
 #    file and build a file list, respectivly.  
 TARFILE=$DATE.tmp
 TARCMD="ssh root@$computer \"$tar --one-file-system -zcf - -C $filesystem . $lastcmd \
	--exclude-from=$exf\" | pee \"cat > $TARFILE\" \"tar tzvf -\""
 echo `date`" $computer: Running $TARCMD" >&2
 eval $TARCMD > $DATE.idx

 if [ -s $TARFILE ]; then
  # File is >0 in size, we assume it worked.

  # Move tar file in place
  mv $TARFILE $DATE.tgz

  # Update timestamp
  echo $DATEs > ../.date && touch -t $DATE ../.date

  # Make a sorted filelist as well
  sort -k6 < $DATE.sdx > $DATE.sdx

  # Fix permissions
  chmod 600 *tgz                      #only for us
  chmod 644 *sdx *idx 2>/dev/null     #everyone can read

 else
  # Something wrong happened.
  rm $TARFILE
  echo `date`" $computer: $TARFILE empty. $backuplevel backup of $computer:$filesystem failed and deleted"

  # We don't want to return 0
  R=1
 fi

}

for computer in $unixcomputers
do
( 
 # Check that the target filesystem is mounted (actually check that it's not
 # the root filesystem)
 rootfilesystem=`df -P /`
 targetfilesystem=`df -P "$storagedir/$computer"`
 if [ "$rootfilesystem" == "$targetfilesystem" ]; then
 	echo `date`" $computer: Target filesystem ($storagedir/$computer) was mounted on /. Aborting"
	continue
 fi

 echo `date`" $computer: Backing up $computer" >&2

 # Try to SSH to the computer without entering a password.
 if ! `ssh -n -o NumberOfPasswordPrompts=0 root@$computer /bin/true`; then
  echo `date`" $computer: Could not use passwordless SSH."

  # We don't want to return 0
  R=1
  # Do next computer
  continue;
 fi
 
 filesystems=""

 # Check dump bit in fstab
 filesystems=`ssh -n root@$computer "cat /etc/fstab" \
  | grep -v nfs \
  | grep -v "^#" \
  | grep -v "^$" \
  | awk '{ if ( $(NF-1) != "0" ) print $2}' `

 # Clean up our dir at this client
 if ! ssh root@$computer "rm -r ~/.backup ; mkdir -m 700 ~/.backup"; then
  echo `date`" $computer: Could not create backup staging area at $computer:~/.backup - skipping backup of $computer"
  R=1
  continue;
 fi

 # Try to copy $exclude to $computer
 if ! scp $exclude root@$computer:~/.backup/exclude > /dev/null; then
  echo `date`" $computer: Could not copy exclude.txt to $computer - skipping backup of $computer"
  R=1
  continue;
 fi

 # Try to copy preeexec and postexec if they exist
# TODO: Gah, clean this mess!
 [ -f $confdir/preexec.$computer ] && (
    scp $confdir/preexec.$computer  root@$computer:~/.backup/preexec >&2 ||
     ( echo `date`" $computer: Could not copy preexec.$computer to $computer:~/.backup/preexec - skipping backup of $computer"
       R=1
       continue
     )
    )
 [ -f $confdir/postexec.$computer ] && (
    scp $confdir/postexec.$computer root@$computer:~/.backup/postexec >&2 ||
     ( echo `date`" $computer: Could not copy postexec.$computer to $computer:~/.backup/postexec - skipping backup of $computer"
       R=1
       continue
     )
    )

 # Try to run preexec if it exist
 if ! ssh root@$computer "[ ! -f ~/.backup/preexec ] || /bin/bash -x ~/.backup/preexec" >&2; then
 	echo `date`" $computer: Could not run $computer:~/.backup/preexec - skipping backup of $computer"
	R=1
	continue
 fi


 for filesystem in $filesystems
 do
  # Remove / and $ (for Windows) for path names
  sfilesystem=`echo $filesystem | tr '\/\$' '__'`

  # Prepare storage area
  mkdir -m 755 -p $storagedir/$computer/$sfilesystem/{full,daglig} 2>/dev/null
  echo $filesystem > ${storagedir}/${computer}/.${sfilesystem}.name 
  chmod 644 ${storagedir}/${computer}/.${sfilesystem}.name

  # Default backuplevel
  backuplevel=daglig

  if [ ! -f $storagedir/$computer/$sfilesystem/.date ]
  then
   # Take the first full backup of this filesystem on this computer
   backuplevel=full
   echo $DATEs > $storagedir/$computer/$sfilesystem/.date
  fi

  # Check if we want a full backup
  if [ -z "`find $storagedir/$computer/$sfilesystem/full/ -name \*tgz -mtime -$dagermellomfulle`" ]; then
   backuplevel=full
  fi
 
  # We want to be here
  cd $storagedir/$computer/$sfilesystem/$backuplevel || die

  # Perform the actual backup
  backup

  # Check if this box has a custom number of full backups
  if [ -f $confdir/maksfulle.$computer ] ; then
    mf=$((`cat $confdir/maksfulle.$computer`+1))
  else
    mf=$(($maksantallfulle+1))
  fi

  # Delete old full backups
  for full in `ls -1t $storagedir/$computer/$sfilesystem/full/*tgz | tail -n +$mf`
  do
   prefix=`echo $full | sed "s/\.[^.]*$//"`
   echo `date`": $computer:$filesystem Deleting full backup $prefix" >&2
   rm $prefix*
  done

  # Delete incremental backups older than the oldest full backup
  oldf=`ls -t1 $storagedir/$computer/$sfilesystem/full/*tgz | tail -1`
  find \
     $storagedir/$computer/$sfilesystem/daglig \
     -type f \
     \! -newer $oldf \
     -printf "`date`: $computer: Deleting old incremental backup: %p\n" \
     -exec rm {} \; >&2
 done

 # Try to run postexec if it exist
 if ! ssh root@$computer "[ ! -f ~/.backup/postexec ] || /bin/bash -x ~/.backup/postexec" >&2; then
 	echo `date`" $computer: Could not run $computer:~/.backup/postexec"
	R=1
 fi

) &
done

wait

# Remove lockfile
rm $LOCKFILE

# Did anything go wrong?
# (IMPORTANT NOTE: The R-business does not work as expected, since this script
# forks itself).
if [ $R != 0 ]; then
	echo `date`": Backup run ended with errors, check logs."
	exit 1
else
	echo `date`": Backup run ended" >&2
fi
